import re

import bslint.error_messages.constants as err_const
import bslint.lexer.regexs as regexs


def find_match(characters):
    token = None
    for token in regexs.REGEXS:
        match = re.match(token.regex, characters, re.IGNORECASE)
        if match:
            break
    if not match:
        raise ValueError(err_const.NO_MATCH_FOUND)
    else:
        return {"match": match, "token_lexer_type": token.lexer_type,
                "token_parser_type": token.parser_type, "indentation_level": token.indentation}



import os

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
TESTS_PATH = os.path.join(ROOT_DIR, 'tests')
TESTS_RESOURCES_PATH = os.path.join(TESTS_PATH, 'resources')
TESTS_CONFIG_PATH = os.path.join(TESTS_RESOURCES_PATH, 'config')
CONFIG_PATH = os.path.join(ROOT_DIR, 'bslint/config')

DEFAULT_CONFIG_FILE_PATH = os.path.join(CONFIG_PATH, 'default-config.json')
TEST_CONFIG_FILE_PATH = os.path.join(TESTS_CONFIG_PATH, 'test-config.json')
BSLINT_COMMAND_CONFIG_PATH = os.path.join(TESTS_CONFIG_PATH, 'bslint_commands')
COMMENTS_CONFIG_PATH = os.path.join(TESTS_CONFIG_PATH, 'comments')

LEXING_TEST_FILES_PATH = os.path.join(TESTS_RESOURCES_PATH, 'lexing_test_files')
STYLING_TEST_FILES_PATH = os.path.join(TESTS_RESOURCES_PATH, 'styling_test_files')
EMPTY_LINES_TEST_FILES_PATH = os.path.join(TESTS_RESOURCES_PATH, 'empty_lines_test_files')
TRACE_TEST_FILES_PATH = os.path.join(TESTS_RESOURCES_PATH, 'trace_test_files')


# pylint: disable=too-few-public-methods
from bslint import constants as const
from bslint.lexer.token import Token


class MatchHandler:
    def __init__(self):
        self._match = None
        self._token_lexer_type = None
        self._token_parser_type = None
        self._type = None

    def match_handler(self, regex_match):
        self._type = None
        self._match = regex_match["match"]
        self._token_lexer_type = regex_match["token_lexer_type"]
        self._token_parser_type = regex_match["token_parser_type"]
        if self._token_lexer_type != const.COMMENT:
            return self._build_token()

    def _build_token(self):
        if self._token_lexer_type == const.ID:
            self._build_id_token()
        elif self._token_lexer_type == const.STRING:
            self._match = self._match.group("value")
        else:
            self._match = self._match.group()
        return Token(self._match, self._token_lexer_type, self._token_parser_type,
                     token_type=self._type)

    def _build_id_token(self):
        if self._match.group('type') is not '':
            self._type = self._match.group('type')
        self._match = self._match.group('value')



# pylint: disable=too-few-public-methods
class Error:

    def __init__(self, message):
        self.message = message

    def get_message(self, params):
        return self.message.format(*(params))



import codecs
import re

import bslint.constants as const
import bslint.error_messages.constants as err_const
import bslint.lexer.regexs as regexs
import bslint.lexer.words_dictionary as words_dict
import bslint.utilities.config_loader as config_loader

COMMENT_REGEX = [regex.regex for regex in regexs.REGEXS if regex.lexer_type == const.COMMENT][0]
DICTIONARY = words_dict.get_new_dictionary()


def check_comment(token):
    if _command_is_active("check_comment") is not True:
        return

    check_comment_config = config_loader.CONFIG["check_comment"]["params"]
    allow_todos = check_comment_config["TODOs"]["allow_TODOs"]
    allow_generic_comments = check_comment_config["allow_generic_comments"]
    todos_format = check_comment_config["TODOs"]["format"]

    if allow_todos and allow_generic_comments:
        if re.match(COMMENT_REGEX + "TODO", token):
            if not re.match(todos_format, token):
                return {"error_key": err_const.NON_CONVENTIONAL_TODO, "error_params": []}

    elif allow_todos and not allow_generic_comments:
        if not re.match(todos_format, token):
            return {"error_key": err_const.NON_CONVENTIONAL_TODO_AND_NO_COMMENTS,
                    "error_params": []}

    elif not allow_todos and allow_generic_comments:
        if re.match(COMMENT_REGEX + "TODO", token, re.IGNORECASE):
            return {"error_key": err_const.NO_TODOS, "error_params": []}

    else:
        return {"error_key": err_const.COMMENTS_NOT_ALLOWED, "error_params": []}
    return


def check_file_encoding(file_path):
    if _command_is_active("check_file_encoding") is not True:
        return

    file_encoding = config_loader.CONFIG['check_file_encoding']['params']
    try:
        codecs.open(file_path, encoding=file_encoding["source_file_encoding"]).read()
    except ValueError:
        return {"error_key": err_const.FILE_ENCODING, "error_params": []}


def check_indentation(current_indentation_level, characters, indentation_level):
    if _command_is_active("check_indentation") is not True:
        return
    if indentation_level == const.DECREMENT_INDENTATION or \
                    indentation_level == const.SPECIAL_INDENTATION:
        current_indentation_level -= 1
    warning = _handle_warnings(current_indentation_level, characters)
    if indentation_level == const.INCREMENT_INDENTATION or \
                    indentation_level == const.SPECIAL_INDENTATION:
        current_indentation_level += 1

    return warning, current_indentation_level


def check_trace_free():
    if _command_is_active("check_trace_free") is True:
        return {"error_key": err_const.TRACEABLE_CODE, "error_params": []}


def check_max_line_length(line_length):
    if _command_is_active("max_line_length") is False:
        return

    max_len = config_loader.CONFIG['max_line_length']['params']['max_line_length']
    if max_len < line_length:
        return {"error_key": err_const.LINE_LENGTH, "error_params": [max_len]}


def check_consecutive_empty_lines(empty_lines):
    if _command_is_active("consecutive_empty_lines") is False:
        return

    params = config_loader.CONFIG["consecutive_empty_lines"]["params"]
    empty_lines_allowed = params["consecutive_empty_lines"]
    if empty_lines > empty_lines_allowed:
        return {"error_key": err_const.CONSECUTIVE_EMPTY_LINES,
                "error_params": [empty_lines_allowed]}


def check_skip_file():
    if _command_is_active("skip_file") is False:
        return
    return True


def check_skip_line(line_number):
    if _command_is_active("skip_line") is False:
        return

    return line_number + 1


def check_spaces_around_operators(characters, current_char_index):
    if _command_is_active("spaces_around_operators") is not True:
        return

    allowed_num_spaces = config_loader.CONFIG["spaces_around_operators"]["params"][
        "spaces_around_operators"]
    before_index = current_char_index - allowed_num_spaces - 2
    after_index = current_char_index + allowed_num_spaces + 1
    chars_around_operator = characters[before_index:after_index]
    if not re.match(r"(\S{0,1})\s{" + str(allowed_num_spaces) + r"}\S\s{" + str(
            allowed_num_spaces) + r"}\S{0,1}$", chars_around_operator):
        return {"error_key": err_const.NO_SPACE_AROUND_OPERATORS,
                "error_params": [allowed_num_spaces]}


def check_spelling(token, token_lexer_type):
    if _command_is_active("spell_check") is not True:
        return

    if token_lexer_type == const.COMMENT:
        words = _parse_comment_words(token)
    else:
        words = _parse_words(token)

    for word in words:
        spelt_correct = DICTIONARY.check(word)
        if not spelt_correct:
            return {"error_key": err_const.TYPO_IN_CODE, "error_params": []}


def check_method_dec_spacing(read_line):
    read_line = read_line.lstrip()

    if _command_is_active("check_method_declaration_spacing") is not True or \
            re.match(COMMENT_REGEX, read_line, re.IGNORECASE):
        return

    method_spaces = config_loader.CONFIG["check_method_declaration_spacing"]["params"][
        "method_spaces"]

    if re.search(r"\b(function|sub)\b", read_line, re.IGNORECASE) and not re.search(
            r"\b(end function|end sub)\b",
            read_line) and not re.search(
                r"\b(endfunction|endsub)\b",
                read_line, re.IGNORECASE):
        if not re.search(r"(function|sub)\s{" + str(
                method_spaces) + r"}[a-z0-9_A-Z]*\(([a-z0-9_A-Z]*" + \
                                 r"(?:(\s{" + str(method_spaces) + r"}as\s{" + str(
                                     method_spaces) + r"}([a-z0-9_A-Z]+))?)(?:,\s{" + \
                                 str(method_spaces) + r"}[a-z0-9_A-Z]*)?)*\)", read_line,
                         re.IGNORECASE):
            return {"error_key": err_const.METHOD_DECLARATION_SPACING, "error_params": []}


# region Private helper functions


def _command_is_active(command_name):
    if command_name in config_loader.CONFIG:
        return config_loader.CONFIG[command_name]["active"]
    else:
        return False


def _change_dict_lang(dict_lang):
    global DICTIONARY
    DICTIONARY = words_dict.get_new_dictionary(dict_lang)


def _handle_warnings(current_indentation_level, characters):
    indent_config = config_loader.CONFIG['check_indentation']['params']
    only_tab_indents = indent_config["only_tab_indents"]
    tab_size = indent_config["tab_size"]
    if re.search(r"\S", characters):
        if only_tab_indents:
            if not re.match(r"\t{" + str(current_indentation_level) + r"}\S", characters):
                return {"error_key": err_const.TAB_AND_SPACES, "error_params": []}
        else:
            if not re.match(r"\s{" + str(
                    tab_size * current_indentation_level) + r"}\S", characters):
                return {"error_key": err_const.TAB_INDENTATION_ERROR,
                        "error_params": [tab_size]}


def _parse_words(identifier_str):
    words = []
    word = ''

    for i in range(0, len(identifier_str)):
        char = identifier_str[i]
        if not char.isalpha():
            if word != '':
                words.append(word)
            word = ''
            continue
        if char.islower():
            word += char
        elif char.isupper():
            if not word == '':
                words.append(word)
            word = ''
            word += char
        elif char == '_' and word != '':
            words.append(word)
            word = ''
    if not word == '':
        words.append(word)

    return words


def _parse_comment_words(comment):
    words = []
    comment_list = comment.split(' ')

    for word in comment_list:
        if not word.isalpha():
            continue
        if word == 'rem':
            continue
        if word.isupper():
            continue
        else:
            if not word == '':
                words.append(word)
    return words

# endregion



import bslint.constants as const
import bslint.parser.rule_builder as rule_builder
RULES_LIST = {
    const.PRIORITY_ZERO: {
        const.ID: [
            ([const.ID, const.DOT, const.ID], const.ID),
            ([const.FUNCTION_CALL, const.DOT, const.ID], const.VALUE)
        ],
        const.CLOSE_SQUARE_BRACKET: [([const.ID, const.OPEN_SQUARE_BRACKET, const.ID, const.CLOSE_SQUARE_BRACKET],
                                      const.ID)],
        const.FUNCTION_CALL: [([const.FUNCTION_CALL, const.DOT, const.FUNCTION_CALL], const.FUNCTION_CALL)],
        const.ELSE: [([const.ELSE], const.ELSE_STMT)],

        # region Closing Statements
        const.END_FOR_TOKEN: [([const.END_FOR_TOKEN], const.END_FOR)],
        const.END_IF_TOKEN: [([const.END_IF_TOKEN], const.END_IF)],
        const.END_WHILE_TOKEN: [([const.END_WHILE_TOKEN], const.END_WHILE)],
        const.END_SUB_TOKEN: [([const.END_SUB_TOKEN], const.END_FUNCTION)],
        const.END_FUNCTION_TOKEN: [([const.END_FUNCTION_TOKEN], const.END_FUNCTION)],
        const.END_TOKEN: [([const.END_TOKEN], const.END)],
        # end region
        const.RETURN: [([const.RETURN], const.RETURN_STMT)],
        const.EXIT: [([const.EXIT], const.EXIT_STMT)],
    },
    const.PRIORITY_ONE: {
        const.ID: [
            ([const.ID, const.OPERATOR, const.ID], const.VALUE),
            ([const.VALUE, const.OPERATOR, const.ID], const.VALUE),
            ([const.FUNCTION_CALL, const.OPERATOR, const.ID], const.VALUE),
            ([const.ID, const.PLUS, const.ID], const.VALUE),
            ([const.ID, const.PLUS, const.PLUS, const.ID], const.VALUE),
            ([const.ID, const.PLUS, const.MINUS, const.ID], const.VALUE),
            ([const.VALUE, const.PLUS, const.ID], const.VALUE),
            ([const.VALUE, const.PLUS, const.PLUS, const.ID], const.VALUE),
            ([const.VALUE, const.PLUS, const.MINUS, const.ID], const.VALUE),
            ([const.FUNCTION_CALL, const.PLUS, const.ID], const.VALUE),
            ([const.FUNCTION_CALL, const.PLUS, const.PLUS, const.ID], const.VALUE),
            ([const.FUNCTION_CALL, const.PLUS, const.MINUS, const.ID], const.VALUE),
            ([const.ID, const.MINUS, const.ID], const.VALUE),
            ([const.ID, const.MINUS, const.MINUS, const.ID], const.VALUE),
            ([const.ID, const.MINUS, const.PLUS, const.ID], const.VALUE),
            ([const.VALUE, const.MINUS, const.ID], const.VALUE),
            ([const.VALUE, const.MINUS, const.MINUS, const.ID], const.VALUE),
            ([const.VALUE, const.MINUS, const.PLUS, const.ID], const.VALUE),
            ([const.FUNCTION_CALL, const.MINUS, const.ID], const.VALUE),
            ([const.FUNCTION_CALL, const.MINUS, const.MINUS, const.ID], const.VALUE),
            ([const.FUNCTION_CALL, const.MINUS, const.PLUS, const.ID], const.VALUE),
            ([const.ID, const.COLON, const.ID], const.ASSOCIATIVE_ARRAY_ARGUMENT),
            ([const.ID, const.EQUALS, const.ID], const.VAR_AS),
            ([const.ID, const.EQUALS, const.MINUS, const.ID], const.VAR_AS),
            ([const.ID, const.EQUALS, const.PLUS, const.ID], const.VAR_AS),
            ([const.FOR, const.VAR_AS, const.TO, const.VALUE, const.STEP, const.ID], const.FOR_STMT),
            ([const.FOR, const.VAR_AS, const.TO, const.ID], const.FOR_STMT),
            ([const.FOR, const.VAR_AS, const.TO, const.ID, const.STEP, const.ID], const.FOR_STMT),
            ([const.FOR, const.VAR_AS, const.TO, const.FUNCTION_CALL, const.STEP, const.ID], const.FOR_STMT),
            ([const.WHILE, const.ID], const.WHILE_STMT),
            ([const.WHILE, const.PLUS, const.ID], const.WHILE_STMT),
            ([const.WHILE, const.MINUS, const.ID], const.WHILE_STMT),
            ([const.PRINT_KEYWORD, const.ID], const.PRINT_STMT),
            ([const.COMMA, const.ID], const.ARGUMENT),
            ([const.SEMI_COLON, const.ID], const.ARGUMENT),
            ([const.FOR_EACH, const.ID, const.IN, const.ID], const.FOR_EACH_STMT),
            ([const.IF, const.ID], const.IF_STMT),
            ([const.VALUE, const.EQUALS, const.ID], const.CONDITION),
            ([const.FUNCTION_CALL, const.EQUALS, const.ID], const.CONDITION),
            ([const.ANONYMOUS_FUNCTION_DECLARATION, const.EQUALS, const.ID], const.CONDITION),
            ([const.VALUE, const.COMPARISON_OPERATOR, const.ID], const.CONDITION),
            ([const.ID, const.COMPARISON_OPERATOR, const.ID], const.CONDITION),
            ([const.FUNCTION_CALL, const.COMPARISON_OPERATOR, const.ID], const.CONDITION),
            ([const.ANONYMOUS_FUNCTION_DECLARATION, const.COMPARISON_OPERATOR, const.ID], const.CONDITION),
            ([const.ELSE_IF, const.ID], const.ELSE_IF_STMT),
            ([const.RETURN_STMT, const.ID], const.RETURN_STMT)
        ],
        const.VALUE: [
            ([const.VALUE, const.OPERATOR, const.VALUE], const.VALUE),
            ([const.ID, const.OPERATOR, const.VALUE], const.VALUE),
            ([const.FUNCTION_CALL, const.OPERATOR, const.VALUE], const.VALUE),
            ([const.VALUE, const.PLUS, const.VALUE], const.VALUE),
            ([const.VALUE, const.PLUS, const.PLUS, const.VALUE], const.VALUE),
            ([const.VALUE, const.PLUS, const.MINUS, const.VALUE], const.VALUE),
            ([const.FUNCTION_CALL, const.PLUS, const.VALUE], const.VALUE),
            ([const.ID, const.PLUS, const.VALUE], const.VALUE),
            ([const.ID, const.PLUS, const.PLUS, const.VALUE], const.VALUE),
            ([const.ID, const.PLUS, const.MINUS, const.VALUE], const.VALUE),
            ([const.FUNCTION_CALL, const.PLUS, const.VALUE], const.VALUE),
            ([const.FUNCTION_CALL, const.PLUS, const.PLUS, const.VALUE], const.VALUE),
            ([const.FUNCTION_CALL, const.PLUS, const.MINUS, const.VALUE], const.VALUE),
            ([const.VALUE, const.MINUS, const.VALUE], const.VALUE),
            ([const.VALUE, const.MINUS, const.MINUS, const.VALUE], const.VALUE),
            ([const.VALUE, const.MINUS, const.PLUS, const.VALUE], const.VALUE),

            ([const.ID, const.MINUS, const.VALUE], const.VALUE),
            ([const.ID, const.MINUS, const.MINUS, const.VALUE], const.VALUE),
            ([const.ID, const.MINUS, const.PLUS, const.VALUE], const.VALUE),
            ([const.FUNCTION_CALL, const.MINUS, const.VALUE], const.VALUE),
            ([const.FUNCTION_CALL, const.MINUS, const.MINUS, const.VALUE], const.VALUE),
            ([const.FUNCTION_CALL, const.MINUS, const.PLUS, const.VALUE], const.VALUE),
            ([const.ID, const.COLON, const.VALUE], const.ASSOCIATIVE_ARRAY_ARGUMENT),
            ([const.ID, const.EQUALS, const.VALUE], const.VAR_AS),

            ([const.ID, const.EQUALS, const.MINUS, const.VALUE], const.VAR_AS),
            ([const.ID, const.EQUALS, const.PLUS, const.VALUE], const.VAR_AS),
            ([const.FOR, const.VAR_AS, const.TO, const.VALUE], const.FOR_STMT),
            ([const.FOR, const.VAR_AS, const.TO, const.VALUE, const.STEP, const.VALUE], const.FOR_STMT),
            ([const.FOR, const.VAR_AS, const.TO, const.ID, const.STEP, const.VALUE], const.FOR_STMT),
            ([const.FOR, const.VAR_AS, const.TO, const.FUNCTION_CALL, const.STEP, const.VALUE], const.FOR_STMT),
            ([const.FOR_EACH, const.ID, const.IN, const.VALUE], const.FOR_EACH_STMT),

            ([const.IF, const.VALUE], const.IF_STMT),
            ([const.IF, const.FUNCTION_CALL, const.EQUALS, const.VALUE], const.IF_STMT),
            ([const.VALUE, const.EQUALS, const.VALUE], const.CONDITION),
            ([const.FUNCTION_CALL, const.EQUALS, const.VALUE], const.CONDITION),
            ([const.ANONYMOUS_FUNCTION_DECLARATION, const.EQUALS, const.VALUE], const.CONDITION),
            ([const.VALUE, const.COMPARISON_OPERATOR, const.VALUE], const.CONDITION),
            ([const.ID, const.COMPARISON_OPERATOR, const.VALUE], const.CONDITION),
            ([const.FUNCTION_CALL, const.COMPARISON_OPERATOR, const.VALUE], const.CONDITION),
            ([const.ANONYMOUS_FUNCTION_DECLARATION, const.COMPARISON_OPERATOR, const.VALUE], const.CONDITION),
            ([const.ELSE_IF, const.VALUE], const.ELSE_IF_STMT),
            ([const.ELSE_IF, const.FUNCTION_CALL, const.EQUALS, const.VALUE], const.ELSE_IF_STMT),
            ([const.RETURN_STMT, const.VALUE], const.RETURN_STMT),
            ([const.WHILE, const.VALUE], const.WHILE_STMT),
            ([const.WHILE, const.PLUS, const.VALUE], const.WHILE_STMT),
            ([const.WHILE, const.MINUS, const.VALUE], const.WHILE_STMT),
            ([const.PRINT_KEYWORD, const.VALUE], const.PRINT_STMT),
            ([const.COMMA, const.VALUE], const.ARGUMENT),
            ([const.SEMI_COLON, const.VALUE], const.ARGUMENT),
        ],
        const.FUNCTION_CALL: [
            ([const.ID, const.OPERATOR, const.FUNCTION_CALL], const.VALUE),
            ([const.VALUE, const.OPERATOR, const.FUNCTION_CALL], const.VALUE),
            ([const.FUNCTION_CALL, const.OPERATOR, const.FUNCTION_CALL], const.VALUE),
            ([const.VALUE, const.PLUS, const.FUNCTION_CALL], const.VALUE),

            ([const.ID, const.PLUS, const.FUNCTION_CALL], const.VALUE),
            ([const.ID, const.PLUS, const.PLUS, const.FUNCTION_CALL], const.VALUE),
            ([const.ID, const.PLUS, const.MINUS, const.FUNCTION_CALL], const.VALUE),

            ([const.VALUE, const.PLUS, const.FUNCTION_CALL], const.VALUE),
            ([const.VALUE, const.PLUS, const.PLUS, const.FUNCTION_CALL], const.VALUE),
            ([const.VALUE, const.PLUS, const.MINUS, const.FUNCTION_CALL], const.VALUE),

            ([const.FUNCTION_CALL, const.PLUS, const.FUNCTION_CALL], const.VALUE),
            ([const.FUNCTION_CALL, const.PLUS, const.PLUS, const.FUNCTION_CALL], const.VALUE),
            ([const.FUNCTION_CALL, const.PLUS, const.MINUS, const.FUNCTION_CALL], const.VALUE),
            ([const.ID, const.MINUS, const.FUNCTION_CALL], const.VALUE),
            ([const.ID, const.MINUS, const.MINUS, const.FUNCTION_CALL], const.VALUE),
            ([const.ID, const.MINUS, const.PLUS, const.FUNCTION_CALL], const.VALUE),

            ([const.VALUE, const.MINUS, const.FUNCTION_CALL], const.VALUE),
            ([const.VALUE, const.MINUS, const.MINUS, const.FUNCTION_CALL], const.VALUE),
            ([const.VALUE, const.MINUS, const.PLUS, const.FUNCTION_CALL], const.VALUE),

            ([const.FUNCTION_CALL, const.MINUS, const.FUNCTION_CALL], const.VALUE),
            ([const.FUNCTION_CALL, const.MINUS, const.MINUS, const.FUNCTION_CALL], const.VALUE),
            ([const.FUNCTION_CALL, const.MINUS, const.PLUS, const.FUNCTION_CALL], const.VALUE),
            ([const.ID, const.COLON, const.FUNCTION_CALL], const.ASSOCIATIVE_ARRAY_ARGUMENT),
            ([const.FUNCTION, const.FUNCTION_CALL], const.FUNCTION_DECLARATION),
            ([const.SUB, const.FUNCTION_CALL], const.FUNCTION_DECLARATION),
            ([const.ID, const.EQUALS, const.FUNCTION_CALL], const.VAR_AS),
            ([const.ID, const.EQUALS, const.MINUS, const.FUNCTION_CALL], const.VAR_AS),
            ([const.ID, const.EQUALS, const.PLUS, const.FUNCTION_CALL], const.VAR_AS),
            ([const.FOR_EACH, const.ID, const.IN, const.FUNCTION_CALL], const.FOR_EACH_STMT),

            ([const.IF, const.FUNCTION_CALL], const.IF_STMT),
            ([const.VALUE, const.EQUALS, const.FUNCTION_CALL], const.CONDITION),
            ([const.FUNCTION_CALL, const.EQUALS, const.FUNCTION_CALL], const.CONDITION),
            ([const.ANONYMOUS_FUNCTION_DECLARATION, const.EQUALS, const.FUNCTION_CALL], const.CONDITION),
            ([const.VALUE, const.COMPARISON_OPERATOR, const.FUNCTION_CALL], const.CONDITION),
            ([const.ID, const.COMPARISON_OPERATOR, const.FUNCTION_CALL], const.CONDITION),
            ([const.FUNCTION_CALL, const.COMPARISON_OPERATOR, const.FUNCTION_CALL], const.CONDITION),
            ([const.ANONYMOUS_FUNCTION_DECLARATION, const.COMPARISON_OPERATOR, const.FUNCTION_CALL], const.CONDITION),
            ([const.ELSE_IF, const.FUNCTION_CALL], const.ELSE_IF_STMT),
            ([const.RETURN_STMT, const.FUNCTION_CALL], const.RETURN_STMT),
            ([const.FOR, const.VAR_AS, const.TO, const.VALUE, const.STEP, const.FUNCTION_CALL], const.FOR_STMT),
            ([const.FOR, const.VAR_AS, const.TO, const.ID, const.STEP, const.FUNCTION_CALL], const.FOR_STMT),
            ([const.FOR, const.VAR_AS, const.TO, const.FUNCTION_CALL], const.FOR_STMT),
            ([const.FOR, const.VAR_AS, const.TO, const.FUNCTION_CALL, const.STEP, const.FUNCTION_CALL], const.FOR_STMT),
            ([const.WHILE, const.FUNCTION_CALL], const.WHILE_STMT),

            ([const.WHILE, const.PLUS, const.FUNCTION_CALL], const.WHILE_STMT),
            ([const.WHILE, const.MINUS, const.FUNCTION_CALL], const.WHILE_STMT),
            ([const.PRINT_KEYWORD, const.FUNCTION_CALL], const.PRINT_STMT),
            ([const.COMMA, const.FUNCTION_CALL], const.ARGUMENT),
            ([const.SEMI_COLON, const.FUNCTION_CALL], const.ARGUMENT),
            ([const.IF_STMT, const.THEN, const.FUNCTION_CALL], const.BLOCK_STMT),
            ([const.ELSE_IF_STMT, const.THEN, const.FUNCTION_CALL], const.BLOCK_STMT),
        ],
        const.TYPE: [
            ([const.FUNCTION, const.FUNCTION_CALL, const.AS, const.TYPE], const.FUNCTION_DECLARATION),
            ([const.ID, const.AS, const.TYPE], const.PARAM),
            ([const.VAR_AS, const.AS, const.TYPE], const.PARAM)
        ],
        const.VAR_AS: [
            ([const.ID, const.DOT, const.VAR_AS], const.VAR_AS),
            ([const.WHILE, const.VAR_AS], const.WHILE_STMT),

            ([const.PRINT_KEYWORD, const.VAR_AS], const.PRINT_STMT),
            ([const.COMMA, const.VAR_AS], const.ARGUMENT),
            ([const.SEMI_COLON, const.VAR_AS], const.ARGUMENT),
            ([const.IF, const.VAR_AS], const.IF_STMT),
            ([const.ELSE_IF, const.VAR_AS], const.ELSE_IF_STMT),
            ([const.RETURN_STMT, const.VAR_AS], const.RETURN_STMT),
            ([const.IF_STMT, const.THEN, const.VAR_AS], const.BLOCK_STMT),
            ([const.ELSE_IF_STMT, const.THEN, const.VAR_AS], const.BLOCK_STMT),
        ],
        const.CLOSE_PARENTHESIS: [
            ([const.FUNCTION, const.OPEN_PARENTHESIS, const.CLOSE_PARENTHESIS], const.ANONYMOUS_FUNCTION_DECLARATION),
            ([const.FUNCTION, const.OPEN_PARENTHESIS, const.PARAM, const.CLOSE_PARENTHESIS],
             const.ANONYMOUS_FUNCTION_DECLARATION),
            ([const.FUNCTION, const.OPEN_PARENTHESIS, const.ID, const.CLOSE_PARENTHESIS],
             const.ANONYMOUS_FUNCTION_DECLARATION),
            ([const.FUNCTION, const.OPEN_PARENTHESIS, const.PARAM, const.ARGUMENT, const.CLOSE_PARENTHESIS],
             const.ANONYMOUS_FUNCTION_DECLARATION),
            ([const.FUNCTION, const.OPEN_PARENTHESIS, const.ID, const.ARGUMENT, const.CLOSE_PARENTHESIS],
             const.ANONYMOUS_FUNCTION_DECLARATION),
            ([const.FUNCTION, const.OPEN_PARENTHESIS, const.VALUE, const.ARGUMENT, const.CLOSE_PARENTHESIS],
             const.ANONYMOUS_FUNCTION_DECLARATION),
            ([const.SUB, const.OPEN_PARENTHESIS, const.CLOSE_PARENTHESIS], const.ANONYMOUS_FUNCTION_DECLARATION),
            ([const.SUB, const.OPEN_PARENTHESIS, const.PARAM, const.CLOSE_PARENTHESIS],
             const.ANONYMOUS_FUNCTION_DECLARATION),
            ([const.SUB, const.OPEN_PARENTHESIS, const.ID, const.CLOSE_PARENTHESIS],
             const.ANONYMOUS_FUNCTION_DECLARATION),
            ([const.SUB, const.OPEN_PARENTHESIS, const.PARAM, const.ARGUMENT, const.CLOSE_PARENTHESIS],
             const.ANONYMOUS_FUNCTION_DECLARATION),
            ([const.SUB, const.OPEN_PARENTHESIS, const.ID, const.ARGUMENT, const.CLOSE_PARENTHESIS],
             const.ANONYMOUS_FUNCTION_DECLARATION),
            ([const.SUB, const.OPEN_PARENTHESIS, const.VALUE, const.ARGUMENT, const.CLOSE_PARENTHESIS],
             const.ANONYMOUS_FUNCTION_DECLARATION),
            ([const.ID, const.OPEN_PARENTHESIS, const.CLOSE_PARENTHESIS], const.FUNCTION_CALL),
            ([const.ID, const.OPEN_PARENTHESIS, const.VALUE, const.CLOSE_PARENTHESIS], const.FUNCTION_CALL),
            ([const.ID, const.OPEN_PARENTHESIS, const.ID, const.CLOSE_PARENTHESIS], const.FUNCTION_CALL),
            ([const.ID, const.OPEN_PARENTHESIS, const.VAR_AS, const.CLOSE_PARENTHESIS], const.FUNCTION_CALL),
            ([const.ID, const.OPEN_PARENTHESIS, const.FUNCTION_CALL, const.CLOSE_PARENTHESIS], const.FUNCTION_CALL),
            ([const.ID, const.OPEN_PARENTHESIS, const.PARAM, const.CLOSE_PARENTHESIS], const.FUNCTION_CALL),
            ([const.ID, const.OPEN_PARENTHESIS, const.VALUE, const.ARGUMENT, const.CLOSE_PARENTHESIS],
             const.FUNCTION_CALL),
            ([const.ID, const.OPEN_PARENTHESIS, const.ID, const.ARGUMENT, const.CLOSE_PARENTHESIS],
             const.FUNCTION_CALL),
            ([const.ID, const.OPEN_PARENTHESIS, const.VAR_AS, const.ARGUMENT, const.CLOSE_PARENTHESIS],
             const.FUNCTION_CALL),
            ([const.ID, const.OPEN_PARENTHESIS, const.FUNCTION_CALL, const.ARGUMENT, const.CLOSE_PARENTHESIS],
             const.FUNCTION_CALL),
            ([const.ID, const.OPEN_PARENTHESIS, const.PARAM, const.ARGUMENT, const.CLOSE_PARENTHESIS],
             const.FUNCTION_CALL),
            ([const.BUILT_IN_FUNCTION, const.OPEN_PARENTHESIS, const.PARAM, const.CLOSE_PARENTHESIS],
             const.FUNCTION_CALL),
            ([const.BUILT_IN_FUNCTION, const.OPEN_PARENTHESIS, const.CLOSE_PARENTHESIS], const.FUNCTION_CALL),
            ([const.BUILT_IN_FUNCTION, const.OPEN_PARENTHESIS, const.VALUE, const.CLOSE_PARENTHESIS],
             const.FUNCTION_CALL),
            ([const.BUILT_IN_FUNCTION, const.OPEN_PARENTHESIS, const.ID, const.CLOSE_PARENTHESIS],
             const.FUNCTION_CALL),
            ([const.BUILT_IN_FUNCTION, const.OPEN_PARENTHESIS, const.FUNCTION_CALL, const.CLOSE_PARENTHESIS],
             const.FUNCTION_CALL),
            ([const.BUILT_IN_FUNCTION, const.OPEN_PARENTHESIS, const.PARAM, const.ARGUMENT, const.CLOSE_PARENTHESIS],
             const.FUNCTION_CALL),
            ([const.BUILT_IN_FUNCTION, const.OPEN_PARENTHESIS, const.VALUE, const.ARGUMENT, const.CLOSE_PARENTHESIS],
             const.FUNCTION_CALL),
            ([const.BUILT_IN_FUNCTION, const.OPEN_PARENTHESIS, const.ID, const.ARGUMENT, const.CLOSE_PARENTHESIS],
             const.FUNCTION_CALL),
            ([const.BUILT_IN_FUNCTION, const.OPEN_PARENTHESIS, const.FUNCTION_CALL, const.ARGUMENT,
              const.CLOSE_PARENTHESIS],
             const.FUNCTION_CALL),
            ([const.OPEN_PARENTHESIS, const.VALUE, const.CLOSE_PARENTHESIS], const.VALUE),
            ([const.OPEN_PARENTHESIS, const.ID, const.CLOSE_PARENTHESIS], const.ID),
            ([const.OPEN_PARENTHESIS, const.VAR_AS, const.CLOSE_PARENTHESIS], const.VAR_AS)
        ],
        const.CLOSE_CURLY_BRACKET: [
            ([const.OPEN_CURLY_BRACKET, const.CLOSE_CURLY_BRACKET], const.ENUMERABLE_OBJECT),
            ([const.OPEN_CURLY_BRACKET, const.ASSOCIATIVE_ARRAY_ARGUMENT, const.CLOSE_CURLY_BRACKET],
             const.ENUMERABLE_OBJECT)
        ],
        const.ASSOCIATIVE_ARRAY_ARGUMENT: [
            ([const.ASSOCIATIVE_ARRAY_ARGUMENT, const.COMMA, const.ASSOCIATIVE_ARRAY_ARGUMENT],
             const.ASSOCIATIVE_ARRAY_ARGUMENT)
        ],
        const.CLOSE_SQUARE_BRACKET: [
            ([const.OPEN_SQUARE_BRACKET, const.CLOSE_SQUARE_BRACKET], const.ENUMERABLE_OBJECT),
            ([const.OPEN_SQUARE_BRACKET, const.VALUE, const.CLOSE_SQUARE_BRACKET], const.ENUMERABLE_OBJECT),
            ([const.OPEN_SQUARE_BRACKET, const.ID, const.CLOSE_SQUARE_BRACKET], const.ENUMERABLE_OBJECT),
            ([const.OPEN_SQUARE_BRACKET, const.VALUE, const.ARGUMENT, const.CLOSE_SQUARE_BRACKET],
             const.ENUMERABLE_OBJECT),
            ([const.OPEN_SQUARE_BRACKET, const.ID, const.ARGUMENT, const.CLOSE_SQUARE_BRACKET],
             const.ENUMERABLE_OBJECT)
        ],
        const.ARRAY_ARGUMENT: [
            ([const.ARRAY_ARGUMENT, const.COMMA, const.ARRAY_ARGUMENT], const.ARRAY_ARGUMENT),
            ([const.ARGUMENT, const.COMMA, const.ARRAY_ARGUMENT], const.ARRAY_ARGUMENT)
        ],
        const.ARGUMENT: [
            ([const.ARRAY_ARGUMENT, const.COMMA, const.ARGUMENT], const.ARRAY_ARGUMENT),
            ([const.ARRAY_ARGUMENT, const.COMMA, const.ARGUMENT], const.ARRAY_ARGUMENT),
            ([const.PRINT_KEYWORD, const.ARGUMENT], const.PRINT_STMT),
            ([const.PRINT_KEYWORD, const.VALUE, const.ARGUMENT], const.PRINT_STMT),
            ([const.PRINT_KEYWORD, const.ID, const.ARGUMENT], const.PRINT_STMT),
            ([const.PRINT_KEYWORD, const.FUNCTION_CALL, const.ARGUMENT], const.PRINT_STMT),
            ([const.PRINT_KEYWORD, const.VAR_AS, const.ARGUMENT], const.PRINT_STMT),
            ([const.PRINT_KEYWORD, const.ARGUMENT, const.ARGUMENT], const.PRINT_STMT),
            ([const.PRINT_KEYWORD, const.ENUMERABLE_OBJECT, const.ARGUMENT], const.PRINT_STMT),
            ([const.COMMA, const.ARGUMENT], const.ARGUMENT),
            ([const.ARGUMENT, const.ARGUMENT], const.ARGUMENT)
        ],
        const.ENUMERABLE_OBJECT: [
            ([const.ID, const.EQUALS, const.ENUMERABLE_OBJECT], const.VAR_AS),
            ([const.PRINT_KEYWORD, const.ENUMERABLE_OBJECT], const.PRINT_STMT),
            ([const.COMMA, const.ENUMERABLE_OBJECT], const.ARGUMENT),
            ([const.SEMI_COLON, const.ENUMERABLE_OBJECT], const.ARGUMENT),
            ([const.RETURN_STMT, const.ENUMERABLE_OBJECT], const.RETURN_STMT)
        ],
        const.PARAM:[
            ([const.PARAM, const.COMMA, const.PARAM], const.PARAM),
            ([const.COMMA, const.PARAM], const.ARGUMENT),
            ([const.SEMI_COLON, const.PARAM], const.ARGUMENT)
        ],
        const.ANONYMOUS_FUNCTION_DECLARATION: [
            ([const.IF, const.ANONYMOUS_FUNCTION_DECLARATION], const.IF_STMT),
            ([const.VALUE, const.EQUALS, const.ANONYMOUS_FUNCTION_DECLARATION], const.CONDITION),
            ([const.FUNCTION_CALL, const.EQUALS, const.ANONYMOUS_FUNCTION_DECLARATION], const.CONDITION),
            ([const.ANONYMOUS_FUNCTION_DECLARATION, const.EQUALS, const.ANONYMOUS_FUNCTION_DECLARATION],
             const.CONDITION),
            ([const.VALUE, const.COMPARISON_OPERATOR, const.ANONYMOUS_FUNCTION_DECLARATION], const.CONDITION),
            ([const.ID, const.COMPARISON_OPERATOR, const.ANONYMOUS_FUNCTION_DECLARATION], const.CONDITION),
            ([const.FUNCTION_CALL, const.COMPARISON_OPERATOR, const.ANONYMOUS_FUNCTION_DECLARATION], const.CONDITION),
            ([const.ANONYMOUS_FUNCTION_DECLARATION, const.COMPARISON_OPERATOR, const.ANONYMOUS_FUNCTION_DECLARATION],
             const.CONDITION),
            ([const.ELSE_IF, const.ANONYMOUS_FUNCTION_DECLARATION], const.ELSE_IF_STMT),
            ([const.RETURN_STMT, const.ANONYMOUS_FUNCTION_DECLARATION], const.RETURN_STMT),
        ],
        const.THEN: [
            ([const.IF, const.CONDITION, const.THEN], const.IF_STMT),
            ([const.IF_STMT, const.THEN], const.IF_STMT),
            ([const.ELSE_IF, const.CONDITION, const.THEN], const.ELSE_IF_STMT),
            ([const.ELSE_IF_STMT, const.THEN], const.ELSE_IF_STMT),
        ],
        const.CONDITION: [
            ([const.IF, const.CONDITION], const.IF_STMT),
            ([const.ELSE_IF, const.CONDITION], const.ELSE_IF_STMT)
        ],
        const.WHILE: [([const.EXIT, const.WHILE], const.EXIT)],
        const.FOR: [([const.EXIT, const.FOR], const.EXIT)],
    }
}

RULES = rule_builder.load_grammar_rules(RULES_LIST)



import re

from bslint import constants as const
from bslint.error_messages import handler as err
from bslint.lexer import commands as commands


class StylingHandler:
    # pylint: disable=too-many-instance-attributes
    def __init__(self, characters):
        self._is_empty_line = True
        self.line_number = 1
        self.line_length = 0
        self._current_indentation_level = 0
        self._indentation_level = 0
        self._token_lexer_type = None
        self._match = None
        self._consecutive_empty_lines = 0
        self.characters = characters
        self.current_char_index = 0
        self.warnings = []
        self._skip_styling_on_file = False
        self.end_of_statement = False
        self.open_curly_braces = 0
        self._line_not_to_style_check = -1

    def _get_last_line(self):
        last_line = re.findall("(?:(?<=^)|(?<=\n))(.*)",
                               self.characters[:self.current_char_index - 1], re.MULTILINE)
        return last_line[-1]

    def apply_bslint_command(self, command_type):
        if command_type == "skip_line":
            self._line_not_to_style_check = commands.check_skip_line(self.line_number)
        elif command_type == "skip_file":
            self._skip_styling_on_file = commands.check_skip_file()

    def apply_styling(self, regex_match):
        self._match = regex_match["match"]
        self._token_lexer_type = regex_match["token_lexer_type"]
        if regex_match["indentation_level"] != const.NO_INDENTATION:
            self._indentation_level = regex_match["indentation_level"]

        applied_common_styling = False
        if self._token_lexer_type == const.NEW_LINE:
            self.end_of_statement = True
            self.apply_new_line_styling()
            self.line_number += 1
            self.line_length = 0
        elif self._token_lexer_type == const.BSLINT_COMMAND:
            self.apply_bslint_command(self._match.group('command'))
        else:
            self.apply_common_styling()
            applied_common_styling = True

        return applied_common_styling

    def style_checking_is_active(self):
        return self.line_number != self._line_not_to_style_check and not self._skip_styling_on_file

    def check_trace_free(self):
        is_trace_free = commands.check_trace_free()
        self._warning_filter(is_trace_free)

    def apply_common_styling(self):
        if not self.style_checking_is_active():
            return
        self._is_empty_line = False
        if self._token_lexer_type == const.COMMENT:
            self._check_comment_styling()
        elif self._token_lexer_type == const.OPERATOR:
            self._check_operator_spacing()
        elif self._token_lexer_type == const.ID:
            self._check_spelling()
        elif self._token_lexer_type == const.PRINT_KEYWORD:
            self.check_trace_free()

    def _check_spelling(self):
        is_spelt_correctly = commands.check_spelling(self._match.group(), self._token_lexer_type)
        self._warning_filter(is_spelt_correctly)

    def _check_operator_spacing(self):
        correct_spacing = commands.check_spaces_around_operators(self.characters,
                                                                 self.current_char_index)
        self._warning_filter(correct_spacing)

    def _check_comment_styling(self):
        is_correct_comment = commands.check_comment(self._match.group())
        self._warning_filter(is_correct_comment)
        self._check_spelling()

    def apply_new_line_styling(self):
        if not self.style_checking_is_active():
            return
        self._count_consecutive_new_lines()
        is_correct_line_length = commands.check_max_line_length(self.line_length)
        self._warning_filter(is_correct_line_length)

        is_consecutive_empty_lines = commands.check_consecutive_empty_lines(
            self._consecutive_empty_lines)
        self._warning_filter(is_consecutive_empty_lines)

        last_read_line = self._get_last_line()
        self._apply_indentation_styling(last_read_line)
        is_correct_method_dec_spacing = commands.check_method_dec_spacing(
            last_read_line)
        self._warning_filter(is_correct_method_dec_spacing)

    def _apply_indentation_styling(self, last_read_line):
        is_correct_indentation = commands.check_indentation(self._current_indentation_level,
                                                            last_read_line,
                                                            self._indentation_level)
        if is_correct_indentation:
            self._warning_filter(is_correct_indentation[0])
            self._current_indentation_level = is_correct_indentation[1]
            self._indentation_level = 0

    def _count_consecutive_new_lines(self):
        if self._is_empty_line is True:
            self._consecutive_empty_lines += 1
        else:
            self._is_empty_line = True
            self._consecutive_empty_lines = 0

    def check_end_of_statement(self):
        if self._token_lexer_type == const.COLON:
            self.end_of_statement = True
        elif self._token_lexer_type == const.OPEN_CURLY_BRACKET:
            self.open_curly_braces += 1
        elif self._token_lexer_type == const.CLOSE_CURLY_BRACKET:
            self.open_curly_braces -= 1

        if self.open_curly_braces is not 0:
            self.end_of_statement = False

    def _warning_filter(self, result):
        if result is not None:
            result["error_params"].append(str(self.line_number))
            warning = err.get_message(result["error_key"], result["error_params"])
            self.warnings += [warning]



import collections


def load_grammar_rules(rules_list):
    grammar_rule = collections.namedtuple('Rule', ['rule', 'result', 'priority'])
    rules = {}
    for priority in rules_list:
        if priority not in rules:
            rules[priority] = {}
        for final_token in rules_list[priority]:
            if final_token not in rules[priority]:
                rules[priority][final_token] = []
            for rule in rules_list[priority][final_token]:
                rules[priority][final_token].append(grammar_rule(rule[0], rule[1], priority))
    return rules



import os
import enchant

from filepaths import CONFIG_PATH


def get_new_dictionary(dictionary_lang="en_GB"):

    personal_words_list_path = os.path.join(CONFIG_PATH, 'personal-words-list.txt')
    return enchant.DictWithPWL(dictionary_lang, personal_words_list_path)



import bslint.constants as const
import bslint.parser.rule_builder as rule_builder

RULES_LIST = {
    const.PRIORITY_ZERO: {
        const.BLOCK_STMT: [
            ([const.BLOCK_STMT, const.BLOCK_STMT], const.BLOCK_STMT),
            ([const.IF_STMT, const.BLOCK_STMT], const.IF_BLOCK),
            ([const.ELSE_IF_STMT, const.BLOCK_STMT], const.ELSE_IF_BLOCK)
        ],
        # region Block Statement reduction RULES
        const.EXIT_STMT: [([const.EXIT_STMT], const.BLOCK_STMT)],
        const.PRINT_STMT: [([const.PRINT_STMT], const.BLOCK_STMT)],
        const.FUNCTION_CALL: [([const.FUNCTION_CALL], const.BLOCK_STMT)],
        const.VAR_AS: [([const.VAR_AS], const.BLOCK_STMT)],
        const.RETURN_STMT: [([const.RETURN_STMT], const.BLOCK_STMT)],
        # endregion

        const.END_WHILE: [
            ([const.WHILE_STMT, const.BLOCK_STMT, const.END_WHILE], const.BLOCK_STMT),
            ([const.WHILE_STMT, const.END_WHILE], const.BLOCK_STMT)
        ],
        const.END_FOR: [
            ([const.FOR_STMT, const.BLOCK_STMT, const.END_FOR], const.BLOCK_STMT),
            ([const.FOR_STMT, const.END_FOR], const.BLOCK_STMT),
            ([const.FOR_EACH_STMT, const.BLOCK_STMT, const.END_FOR], const.BLOCK_STMT),
            ([const.FOR_EACH_STMT, const.END_FOR], const.BLOCK_STMT),
        ],

        const.END_FUNCTION: [
            ([const.FUNCTION_DECLARATION, const.BLOCK_STMT, const.END_FUNCTION], const.BLOCK_STMT),
            ([const.FUNCTION_DECLARATION, const.END_FUNCTION], const.BLOCK_STMT)
        ],

        # region IF statement
        const.END_IF: [
            ([const.IF_STMT, const.BLOCK_STMT, const.END_IF], const.BLOCK_STMT),
            ([const.ELSE_STMT, const.BLOCK_STMT, const.END_IF], const.END_IF),
            ([const.IF_BLOCK, const.ELSE_IF_BLOCK, const.END_IF], const.BLOCK_STMT)
        ],
        const.ELSE_IF_BLOCK: [([const.ELSE_IF_BLOCK, const.ELSE_IF_BLOCK], const.ELSE_IF_BLOCK)],
        # endregion
    }
}

RULES = rule_builder.load_grammar_rules(RULES_LIST)



# pylint: disable=too-many-instance-attributes
import os
import re
import sys

import bslint
import bslint.constants as const
import bslint.lexer.commands as commands
from bslint.lexer.lexer import Lexer as Lexer


class InterfaceHandler:
    def __init__(self, out=sys.stdout):
        self.is_lexed_correctly = True
        self.files = []
        self.messages = {"Errors": {}, "Warnings": {}}
        self.bslintrc = {"ignore": []}
        self.out = out
        self.manifest_path = ""
        self.warnings_total = 0
        self.errors_total = 0

    def main(self):
        print(const.TITLE_COLOUR + "BSLint: A linter for BrightScript. %s. \n" %
              InterfaceHandler.get_version() + const.END_COLOUR)

        if self.is_specific_path():
            filename = sys.argv[1]
            if not os.path.exists(filename):
                self.is_lexed_correctly = False
                self.out.write(
                    const.ERROR_COLOUR + "The path you have provided does not exist." + const.END_COLOUR + "\n")
                return

        self.manifest_path = self._get_manifest_path()
        self.bslintrc = self._parse_bslintrc(self.manifest_path)

        if self.bslintrc is None:
            return

        self.out.write("\n")

        if self.is_specific_path():
            if os.path.isfile(filename):
                self.lint_file(filename)
            else:
                self.lint_all(filename)
        else:
            pathname = os.getcwd()
            self.lint_all(pathname)
        self.out.write(
            const.FILE_COLOUR + "LINTING COMPLETE" + const.END_COLOUR + "\n")
        if self.is_lexed_correctly:
            print(const.PASS_COLOUR + 'All linted correctly' + const.END_COLOUR)
            print("\n")
        else:
            self.out.write(const.TOTAL_COLOUR + "TOTAL WARNINGS: " + str(self.warnings_total) + const.END_COLOUR + "\n")
            self.out.write(const.TOTAL_COLOUR + "TOTAL ERRORS: " + str(self.errors_total) + const.END_COLOUR + "\n")

    def _get_manifest_path(self):
        try:
            if self.is_specific_path():
                if os.path.isfile(sys.argv[1]):
                    upper_dir = os.path.dirname(sys.argv[1])
                else:
                    upper_dir = sys.argv[1]
            else:
                upper_dir = ""
            count = 0
            while self.no_manifest_in_folder(upper_dir) and count < 5:
                upper_dir = os.path.join(upper_dir, "../")
                count += 1
            if count == 5:
                raise FileNotFoundError
            return upper_dir
        except FileNotFoundError:
            print(const.ERROR_COLOUR + "No manifest file found" + const.END_COLOUR)
            return ""

    def lint_all(self, directory):
        for dir_name, subdirList, files in os.walk(directory):  # pylint: disable=C0103, W0612
            relative_path = self.get_relative_path(dir_name)
            if "ignore" in self.bslintrc and not self.ignore_dir(relative_path, self.bslintrc["ignore"]):
                self.lint_directory(dir_name, files)
        self.print_errors()

    def lint_directory(self, dir_name, files):
        for file in files:
            if file.endswith(".brs") or file.endswith(".bs"):
                filepath = os.path.join(dir_name, file)
                self.lint_file(filepath)

    def lint_file(self, filepath):
        filename = filepath.replace(os.getcwd() + '/', '')
        if filename.endswith(".brs") or filename.endswith(".bs"):
            self.files.append(filename)
            read_file = self.file_reader(filename)
            lex_result = Lexer().lex(read_file['str_to_lex'])
            if lex_result["Status"] == "Error":
                self.handle_lexing_error(filepath, lex_result)
            elif lex_result["Warnings"]:
                self.handle_lexing_warnings(filepath, lex_result)
        else:
            self.is_lexed_correctly = False
        self.print_warnings(filepath)

    def handle_lexing_warnings(self, filepath, lex_result):
        self.is_lexed_correctly = False
        self.messages["Warnings"][filepath] = []
        for warning in lex_result["Warnings"]:
            self.messages["Warnings"][filepath].append(
                const.WARNING_COLOUR + str(warning) + const.END_COLOUR)

    def handle_lexing_error(self, filepath, lex_result):
        self.is_lexed_correctly = False
        self.messages["Errors"][filepath] = []
        for error in lex_result["Tokens"]:
            self.messages["Errors"][filepath].append(
                const.ERROR_COLOUR + str(error) + const.END_COLOUR)

    def print_warnings(self, file_name):
        if file_name in self.messages["Warnings"]:
            self.out.write(const.FILE_COLOUR + file_name + const.END_COLOUR + "\n")
            for message in self.messages["Warnings"][file_name]:
                self.out.write(message + "\n")
            number_warnings = len(self.messages["Warnings"][file_name])
            self.warnings_total += number_warnings
            self.out.write(const.TOTAL_COLOUR + "WARNINGS IN FILE: " + str(number_warnings) + const.END_COLOUR + "\n")
            self.out.write("\n")

    def print_errors(self):
        for file_name in self.messages["Errors"]:
            self.out.write(const.FILE_COLOUR + file_name + const.END_COLOUR + "\n")
            for message in self.messages["Errors"][file_name]:
                self.out.write(message + "\n")
            number_errors = len(self.messages["Errors"][file_name])
            self.errors_total += number_errors
            self.out.write(const.TOTAL_COLOUR + "ERRORS IN FILE: " + str(number_errors) + const.END_COLOUR + "\n")
            self.out.write("\n")

    def ignore_dir(self, relative_directory_path, directories_to_ignore):
        for directory in directories_to_ignore:
            directory = self.add_slashes_to_dir(directory)
            relative_directory_path = self.add_slashes_to_dir(relative_directory_path)

            if relative_directory_path.startswith(directory):
                return True
        return False

    @staticmethod
    def _parse_bslintrc(manifest_path):
        bslintrc_path = os.path.join(manifest_path, ".bslintrc")
        return bslint.config_loader.load_config_file(bslintrc_path)

    def get_relative_path(self, dir_name):
        directory = os.path.abspath(self.manifest_path)
        return dir_name.replace(directory, '')

    @staticmethod
    def add_slashes_to_dir(directory):
        if not directory.startswith("/"):
            directory = "/" + directory
        if not directory.endswith("/"):
            directory += "/"
        return directory

    @staticmethod
    def file_reader(file_to_lex):
        file = open(file_to_lex, "r+")
        str_to_lex = file.read()
        return {"invalid_encoding": commands.check_file_encoding(file_to_lex),
                "str_to_lex": str_to_lex}

    @staticmethod
    def get_version():
        this_dir = os.path.dirname(__file__)
        return re.search(
            r'^__version__\s*=\s*"(.*)"',
            open(os.path.join(this_dir, 'bslint.py')).read(),
            re.M
        ).group(1)

    @staticmethod
    def is_specific_path():
        return len(sys.argv) != 1

    @staticmethod
    def no_manifest_in_folder(upper_dir):
        return not os.path.exists(os.path.join(upper_dir, "MANIFEST")) and not os.path.exists(
            os.path.join(upper_dir, "manifest"))



import bslint.error_messages.builder as err_builder

ERROR = "ERROR"
WARNING = "WARNING"

LINE_LENGTH = "LINE_LENGTH"
FILE_ENCODING = "FILE_ENCODING"
CONSECUTIVE_EMPTY_LINES = "CONSECUTIVE_EMPTY_LINES"
UPPERCASE_VARIABLE = "UPPERCASE_VARIABLE"
TAB_INDENTATION_ERROR = "TAB_INDENTATION_ERROR"
TAB_AND_SPACES = "TAB_AND_SPACES"
TYPO_IN_CODE = "TYPO_IN_CODE"
TRACEABLE_CODE = "TRACEABLE_CODE"
NON_CONVENTIONAL_TODO = "NON_CONVENTIONAL_TODO"
COMMENTS_NOT_ALLOWED = "COMMENTS_NOT_ALLOWED"
NO_RETURN_AND_PARAMETERS_TYPES = "NO_RETURN_AND_PARAMETERS_TYPES"
NOT_CAPITALISED_VARIABLE_NAMES = "NOT_CAPITALISED_VARIABLE_NAMES"
NO_SPACE_AROUND_OPERATORS = "NO_SPACE_AROUND_OPERATORS = {NO_SPACE_AROUND_OPERATORS"
NO_SPACE_BETWEEN_METHOD_AND_PARAMETERS = "NO_SPACE_BETWEEN_METHOD_AND_PARAMETERS"
UNMATCHED_QUOTATION_MARK = "UNMATCHED_QUOTATION_MARK"
NON_CONVENTIONAL_TODO_AND_NO_COMMENTS = "NON_CONVENTIONAL_TODO_AND_NO_COMMENTS"
NO_TODOS = "NO_TODOS"
METHOD_DECLARATION_SPACING = "METHOD_DECLARATION_SPACING"
UNMATCHED_TOKEN = "UNMATCHED_TOKEN"
NO_MATCH_FOUND = "NO_MATCH_FOUND"
STMT_PARSING_FAILED = "STMT_PARSING_FAILED"
PROGRAM_PARSING_FAILED = "PROGRAM_PARSING_FAILED"
NO_SUCH_KEY = "NO_SUCH_KEY"

ERROR_TABLE = {
    LINE_LENGTH: err_builder.Error(
        WARNING + ": Line length exceeds {} number of characters. Line number: {}"),
    FILE_ENCODING: err_builder.Error(WARNING + ": Your files should conform to {} encoding"),
    CONSECUTIVE_EMPTY_LINES: err_builder.Error(
        WARNING + ": Cannot have more than {} consecutive empty lines. Line number: {}"),
    UPPERCASE_VARIABLE: err_builder.Error(
        WARNING + ": Variables should begin with a lowercase letter: line number: {}"),
    TAB_INDENTATION_ERROR: err_builder.Error(
        WARNING + ": Tab indentation should be set to {} spaces. Line number: {}"),
    TAB_AND_SPACES: err_builder.Error(
        WARNING + ": Invalid indentation, you must indent with tabs. Line number: {}"),
    TYPO_IN_CODE: err_builder.Error(
        WARNING + ": You have spelling mistakes in your code. Line number: {}"),
    TRACEABLE_CODE: err_builder.Error(WARNING + ": Print statements not allowed. Line number: {}"),
    NON_CONVENTIONAL_TODO_AND_NO_COMMENTS: err_builder.Error(
        WARNING + ": comments must be TODOs and must follow convention. Line number: {}"),
    COMMENTS_NOT_ALLOWED: err_builder.Error(WARNING + ": No comments allowed. Line number: {}"),
    NO_RETURN_AND_PARAMETERS_TYPES: err_builder.Error(
        WARNING + ": No return and parameter types declared. Line number: {}"),
    NOT_CAPITALISED_VARIABLE_NAMES: err_builder.Error(
        WARNING + ": Variable types must be capitalised. Line number: {}"),
    NO_SPACE_AROUND_OPERATORS: err_builder.Error(
        WARNING + ": Operators must be surrounded by {} space(s). Line number: {}"),
    NO_SPACE_BETWEEN_METHOD_AND_PARAMETERS: err_builder.Error(
        WARNING + ": You must put a space between method name and parameter list. Line number: {}"),
    UNMATCHED_QUOTATION_MARK: err_builder.Error(
        ERROR + ": You have unmatched quotation marks at {} on line number: {}"),
    NON_CONVENTIONAL_TODO: err_builder.Error(
        WARNING + ": TODOs must follow convention. Line number: {}"),
    NO_TODOS: err_builder.Error(WARNING + ": comments must not be TODOs. Line number: {}"),
    METHOD_DECLARATION_SPACING: err_builder.Error(
        WARNING + ": The spacing on the method declaration is incorrect: {}"),
    UNMATCHED_TOKEN: err_builder.Error(ERROR + ": Unmatched tokens, expected {}, but got {}"),
    STMT_PARSING_FAILED: err_builder.Error(ERROR + ": Parsing statement failed. Line number: {}"),
    PROGRAM_PARSING_FAILED: err_builder.Error(ERROR + ": Parsing program failed. Line number: {}")
}



import bslint.error_messages.constants as err_const


def get_message(key, params=None):
    params = params or []
    if key not in err_const.ERROR_TABLE:
        raise ValueError(err_const.NO_SUCH_KEY)
    return err_const.ERROR_TABLE[key].get_message(params)



# pylint: disable=too-many-instance-attributes
import re

import bslint.error_messages.constants as err_const
import bslint.error_messages.handler as err
import bslint.lexer.handlers.match_handler as match_handler
import bslint.lexer.handlers.regex_handler as regex_handler
import bslint.lexer.handlers.styling_handler as styling_handler
from bslint import constants as const


class Lexer:
    def __init__(self):
        self.tokens = []
        self.errors = []
        self.handle_match = match_handler.MatchHandler()
        self.characters = ""
        self.handle_style = None
        self.preceding_token = None
        self.is_valid_token = True
        self.parsing_failed = False
        self.current_token_index = 0
        self.statements_counter = 0

    def lex(self, characters):
        self.characters = characters
        self.handle_style = styling_handler.StylingHandler(characters)
        while self.handle_style.current_char_index < len(self.characters):
            try:
                self.create_token_and_handle_styling()
            except ValueError as exception:
                self.parsing_failed = True
                exception_code = self.get_exception_code(exception)
                if exception_code == err_const.STMT_PARSING_FAILED:
                    self.handle_parsing_error(exception_code)
                    break
                else:
                    self.handle_unexpected_token()
        self.handle_style.apply_new_line_styling()
        if not self.parsing_failed:
            try:
                self.check_statement_validity(self.tokens[self.current_token_index:])
                self.check_program_validity()
            except ValueError as exception:
                self.parsing_failed = True
                self.handle_parsing_error(self.get_exception_code(exception))

        if len(self.errors) is not 0 or self.parsing_failed:
            return {"Status": "Error", "Tokens": self.errors, "Warnings": self.handle_style.warnings}
        else:
            return {"Status": "Success", "Tokens": self.tokens, "Warnings": self.handle_style.warnings}

    def create_token_and_handle_styling(self):
        regex_match = regex_handler.find_match(self.characters[self.handle_style.current_char_index:])
        self.handle_style.line_length += len(regex_match["match"].group())
        self.handle_style.current_char_index += len(regex_match["match"].group())
        if regex_match["token_lexer_type"] is not None:
            applied_common_styling = self.handle_style.apply_styling(regex_match)
            if applied_common_styling:
                token = self.handle_match.match_handler(regex_match)
                if token is not None:
                    token.line_number = self.handle_style.line_number
                    self.tokens.append(token)
            self.handle_style.check_end_of_statement()
            if self.handle_style.end_of_statement is True:
                self.check_statement_validity(self.tokens[self.current_token_index:])
                self.handle_style.end_of_statement = False
                self.current_token_index = len(self.tokens)

    def handle_unexpected_token(self):
        end_of_line = re.match(r"(.*)\n", self.characters[self.handle_style.current_char_index:])
        if end_of_line is None:
            end_of_line = re.match(r"(.*)$", self.characters[self.handle_style.current_char_index:])
        self.errors.append(err.get_message(err_const.UNMATCHED_QUOTATION_MARK,
                                           [(end_of_line.group()[:const.PENULTIMATE_CHARACTER]),
                                            self.handle_style.line_number]))
        self.handle_style.line_number += 1
        self.handle_style.current_char_index += len(end_of_line.group())

    # pylint: disable=unused-argument
    def check_statement_validity(self, statement):
        self.statements_counter += 1

    # pylint: disable=no-self-use
    def check_program_validity(self):
        return

    # pylint: disable=no-self-use
    def handle_parsing_error(self, exception):
        return

    @staticmethod
    def get_exception_code(exception):
        return exception.args[0]



from bslint.parser import statement_reduction_rules as statement_grammar
from bslint.parser import program_reduction_rules as program_grammar
from bslint.error_messages import constants as err_const
from bslint import constants as const
from bslint.lexer.lexer import Lexer
from bslint.error_messages import handler as err

MAX_FINAL_STATEMENT_LENGTH = 1
INVALID_STATEMENT = [const.CONDITION]


class Parser(Lexer):
    # pylint: disable=too-many-instance-attributes
    def __init__(self):
        Lexer.__init__(self)
        self.expected_statement = None
        self.all_statements = []
        self.current_tokens = None
        self.program = []
        self.line_reductions = []
        self.current_output_list = self.all_statements
        self.current_grammar = statement_grammar
        self.number_of_priorities = None
        self.current_priority_level = 0

    def parse(self, characters):
        self.set_number_of_priorities_level()
        return Lexer.lex(self, characters)

    def set_number_of_priorities_level(self):
        self.number_of_priorities = len(self.current_grammar.RULES.keys())

    def check_statement_validity(self, statement):
        self.current_tokens = self._get_token_types(statement)
        if len(self.current_tokens) > 0:
            self.iterate_over_statement()
            if self.current_tokens[0] in INVALID_STATEMENT:
                raise ValueError(err_const.STMT_PARSING_FAILED)
            else:
                self.program.append(self.current_tokens[0])

    @staticmethod
    def _get_token_types(tokens):
        return [token.parser_type for token in tokens]

    def iterate_over_statement(self):
        self.current_priority_level = 0
        while self.current_priority_level < self.number_of_priorities:
            self.reduce_and_handle_error()
            self.current_priority_level += 1
            if len(self.current_tokens) == MAX_FINAL_STATEMENT_LENGTH:
                break

    def reduce_and_handle_error(self):
        is_valid_statement = False
        index = 0
        while index < len(self.current_tokens):
            possible_production_rules = self._get_possible_production_rules(
                self.current_tokens[-(index + 1)])
            matching_production = self._find_matching_production(index, possible_production_rules)
            if matching_production is not None:
                self._replace_current_tokens(index, matching_production)
                self.current_priority_level = 0
                if len(self.current_tokens) > MAX_FINAL_STATEMENT_LENGTH:
                    self.reduce_and_handle_error()
                is_valid_statement = True
            index += 1
        if is_valid_statement is False and self.current_priority_level == self.number_of_priorities - 1:
            raise ValueError(err_const.STMT_PARSING_FAILED)

    def _find_matching_production(self, index, possible_production_rules):
        matching_production = None
        for production in possible_production_rules:
            number_of_tokens_from_end = self._get_number_of_tokens_from_end(production, index)
            current_tokens = self._get_current_tokens(index, number_of_tokens_from_end)
            if production.rule == current_tokens:
                matching_production = production
                break
        return matching_production

    def _replace_current_tokens(self, index, rule):
        tokens_from_end = self._get_number_of_tokens_from_end(rule, index)
        if index == 0:
            del self.current_tokens[-tokens_from_end:]
        else:
            del self.current_tokens[-tokens_from_end: -index]
        self.current_tokens[
            len(self.current_tokens) - index:len(self.current_tokens) - index] = [rule.result]
        self.current_output_list.append(self.current_tokens[:])

    def _get_current_tokens(self, index, tokens_from_end):
        if index == 0:
            current_tokens = self.current_tokens[-tokens_from_end:]
        else:
            current_tokens = self.current_tokens[-tokens_from_end: -index]
        return current_tokens

    @staticmethod
    def _get_number_of_tokens_from_end(production, index):
        return len(production.rule) + index

    def _get_possible_production_rules(self, current_token):
        if current_token not in self.current_grammar.RULES[self.current_priority_level]:
            return []
        else:
            return self.current_grammar.RULES[self.current_priority_level][current_token]

    def check_program_validity(self):
        self.current_grammar = program_grammar
        self.current_tokens = self.program
        self.current_output_list = self.line_reductions
        self.set_number_of_priorities_level()
        try:
            self.iterate_over_statement()
        except ValueError:
            raise ValueError(err_const.PROGRAM_PARSING_FAILED)

    def handle_parsing_error(self, err_code):
        self.errors.append(err.get_message(err_code, [self.handle_style.line_number]))



# -*- coding: utf-8 -*-
"""
BSLint
-----
BSLint is a linter for the BrightScript language.
"""

import re
from setuptools import setup, find_packages

version = re.search(
    '^__version__\s*=\s*"(.*)"',
    open('bslint/bslint.py').read(),
    re.M
).group(1)

setup(
    name="bslint",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    package_data={'bslint': ['config/*.json', 'config/personal-words-list.txt', '../filepaths.py']},
    entry_points={"console_scripts": ['bslint = bslint.bslint:main']},
    version=version,
    description="A linter tool for the BrightScript language.",
    author="BSLint",
    author_email="zachary.robinson@sky.uk",
    url="https://github.com/sky-uk/bslint",
    install_requires=['pyenchant==1.6.8']
)



STMT = 'STMT'
ID = 'ID'
STRING = 'STRING'
INT = 'INT'
NUMERIC = 'NUMERIC'
NEW_LINE = "NEW_LINE"
OPERATOR = "OPERATOR"
COMPARISON_OPERATOR = "COMPARISON_OPERATOR"
SPECIAL_OPERATOR = "SPECIAL_OPERATOR"
BRACKET = "BRACKET"
SQUARE_BRACKET = "SQUARE_BRACKET"
KEYWORD = "KEYWORD"
PRINT_KEYWORD = "PRINT_KEYWORD"
BSLINT_COMMAND = 'BSLINT_COMMAND'
COMMENT = "COMMENT"
ERROR = "ERROR"
WARNING = "WARNING"
CURLY_BRACKET = 'CURLY_BRACKET'
SEMI_COLON = 'SEMI_COLON'
AT_SYMBOL = 'AT_SYMBOL'
HASH_SYMBOL = 'HASH_SYMBOL'
VALUE = "VALUE"
NOT = "NOT"
WHILE = "WHILE"
FOR = "FOR"
FOR_EACH = "FOR_EACH"
TYPE = "TYPE"
DOT = "DOT"
COLON = "COLON"
AS = "AS"
AND = "AND"
OR = "OR"
COMMA = "COMMA"
OPEN_PARENTHESIS = "OPEN_PARENTHESIS"
CLOSE_PARENTHESIS = "CLOSE_PARENTHESIS"
OPEN_SQUARE_BRACKET = "OPEN_SQUARE_BRACKET"
CLOSE_SQUARE_BRACKET = "CLOSE_SQUARE_BRACKET"
OPEN_CURLY_BRACKET = "OPEN_CURLY_BRACKET"
CLOSE_CURLY_BRACKET = "CLOSE_CURLY_BRACKET"
PLUS = 'PLUS'
MINUS = 'MINUS'
EQUALS = "EQUALS"
INVALID = "INVALID"
NATIVE_FUNCTION = "NATIVE_FUNCTION"
COMPONENT = "COMPONENT"
INTERFACE = "INTERFACE"
EVENT = "EVENT"
FUNCTION = "FUNCTION"
SUB = "SUB"
IN = "IN"
TO = "TO"
STEP = "STEP"
IF_STMT = "IF_STMT"
IF = "IF"
THEN = "THEN"
EXIT = "EXIT"
GOTO = "GOTO"
RETURN = "RETURN"
RETURN_STMT = "RETURN_STMT"
VAR_AS = "VAR_AS"
FUNCTION_CALL = "FUNCTION_CALL"
BUILT_IN_FUNCTION = "BUILT_IN_FUNCTION"
ANONYMOUS_FUNCTION_DECLARATION = "ANONYMOUS_FUNCTION_DECLARATION"
ARGUMENT = "ARGUMENT"
FUNCTION_DECLARATION = "FUNCTION_DECLARATION"
PARAM = "PARAM"
FOR_STMT = "FOR_STMT"
FOR_EACH_STMT = "FOR_EACH_STMT"
WHILE_STMT = "WHILE_STMT"
PRINT_STMT = "PRINT_STMT"
BLOCK_STMT = "BLOCK_STMT"
ELSE_IF = "ELSE_IF"
ELSE_IF_STMT = "ELSE_IF_STMT"
ELSE_IF_BLOCK = "ELSE_IF_BLOCK"
IF_BLOCK = "IF_BLOCK"
ELSE = "ELSE"
ELSE_STMT = "ELSE_STMT"
ENUMERABLE_OBJECT = "ENUMERABLE_OBJECT"
ARRAY_ARGUMENT = "ARRAY_ARGUMENT"
ASSOCIATIVE_ARRAY_ARGUMENT = "ASSOCIATIVE_ARRAY_ARGUMENT"
CONDITION = "CONDITION"
EXIT_STMT = "EXIT_STMT"

END_IF = "END_IF"
END_FOR = "END_FOR"
END_WHILE = "END_WHILE"
END_FUNCTION = "END_FUNCTION"
END_SUB = "END_SUB"
END = "END"
END_IF_TOKEN = "END_IF_TOKEN"
END_FOR_TOKEN = "END_FOR_TOKEN"
END_WHILE_TOKEN = "END_WHILE_TOKEN"
END_FUNCTION_TOKEN = "END_FUNCTION_TOKEN"
END_SUB_TOKEN = "END_SUB_TOKEN"
END_TOKEN = "END_TOKEN"

INCREMENT_INDENTATION = 1
DECREMENT_INDENTATION = -1
NO_INDENTATION = 0
SPECIAL_INDENTATION = 2
PENULTIMATE_CHARACTER = -2
WARNING_COLOUR = '\033[93m'
PASS_COLOUR = '\033[92m'
END_COLOUR = '\033[0m'
ERROR_COLOUR = '\033[91m'
FILE_COLOUR = '\033[34;4m'
TITLE_COLOUR = '\033[95m'
HEADER_COLOUR = '\033[34;4m'

PRIORITY_ZERO = 0
PRIORITY_ONE = 1
TOTAL_COLOUR = '\033[35m'



# this characters need escaping . ^ $ * + ? { } [ ] \ | ( )
import collections
import bslint.constants as const

Regex = collections.namedtuple('Token', ['regex', 'lexer_type', 'parser_type', 'indentation'])

REGEX_LIST = [
    [r"\n", const.NEW_LINE, const.NEW_LINE, const.NO_INDENTATION],
    [r"\s", None, None, const.NO_INDENTATION],

    [r"\+=", const.OPERATOR, const.OPERATOR, const.NO_INDENTATION],
    [r"-=", const.OPERATOR, const.OPERATOR, const.NO_INDENTATION],
    [r"\*=", const.OPERATOR, const.OPERATOR, const.NO_INDENTATION],
    [r"/=", const.OPERATOR, const.OPERATOR, const.NO_INDENTATION],
    [r"\\=", const.OPERATOR, const.OPERATOR, const.NO_INDENTATION],  # divide integer
    ["<<=", const.OPERATOR, const.OPERATOR, const.NO_INDENTATION],
    [">>=", const.OPERATOR, const.OPERATOR, const.NO_INDENTATION],
    ["<>", const.OPERATOR, const.COMPARISON_OPERATOR, const.NO_INDENTATION],
    [">=", const.OPERATOR, const.COMPARISON_OPERATOR, const.NO_INDENTATION],
    ["<=", const.OPERATOR, const.COMPARISON_OPERATOR, const.NO_INDENTATION],
    [">=", const.OPERATOR, const.COMPARISON_OPERATOR, const.NO_INDENTATION],
    ["=>", const.OPERATOR, const.COMPARISON_OPERATOR, const.NO_INDENTATION],
    ["=<", const.OPERATOR, const.COMPARISON_OPERATOR, const.NO_INDENTATION],

    [r">>", const.OPERATOR, const.OPERATOR, const.NO_INDENTATION],
    [r"<<", const.OPERATOR, const.OPERATOR, const.NO_INDENTATION],

    [r";", const.SEMI_COLON, const.SEMI_COLON, const.NO_INDENTATION],
    [r"@", const.AT_SYMBOL, const.AT_SYMBOL, const.NO_INDENTATION],
    [r"#", const.HASH_SYMBOL, const.HASH_SYMBOL, const.NO_INDENTATION],
    [r"\.", const.SPECIAL_OPERATOR, const.DOT, const.NO_INDENTATION],
    [r":", const.COLON, const.COLON, const.NO_INDENTATION],
    [r"\^", const.OPERATOR, const.OPERATOR, const.NO_INDENTATION],
    [r"\&", const.OPERATOR, const.OPERATOR, const.NO_INDENTATION],
    [r"\\", const.OPERATOR, const.OPERATOR, const.NO_INDENTATION],
    [r"=", const.OPERATOR, const.EQUALS, const.NO_INDENTATION],
    [r"-", const.OPERATOR, const.MINUS, const.NO_INDENTATION],
    [r"\+", const.OPERATOR, const.PLUS, const.NO_INDENTATION],
    [r"\*", const.OPERATOR, const.OPERATOR, const.NO_INDENTATION],
    [r"/", const.OPERATOR, const.OPERATOR, const.NO_INDENTATION],
    ["<", const.OPERATOR, const.COMPARISON_OPERATOR, const.NO_INDENTATION],
    [">", const.OPERATOR, const.COMPARISON_OPERATOR, const.NO_INDENTATION],

    [r"\(", const.BRACKET, const.OPEN_PARENTHESIS, const.NO_INDENTATION],
    [r"\)", const.BRACKET, const.CLOSE_PARENTHESIS, const.NO_INDENTATION],
    [r"\[", const.SQUARE_BRACKET, const.OPEN_SQUARE_BRACKET, const.NO_INDENTATION],
    [r"\]", const.SQUARE_BRACKET, const.CLOSE_SQUARE_BRACKET, const.NO_INDENTATION],
    [r"\{", const.OPEN_CURLY_BRACKET, const.OPEN_CURLY_BRACKET, const.NO_INDENTATION],
    [r"\}", const.CLOSE_CURLY_BRACKET, const.CLOSE_CURLY_BRACKET, const.NO_INDENTATION],
    [r",", const.SPECIAL_OPERATOR, const.COMMA, const.NO_INDENTATION],
    [r"(TRUE)\b", const.KEYWORD, const.VALUE, const.NO_INDENTATION],
    [r"(GETLASTRUNCOMPILEERROR)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(GETLASTRUNRUNTIMEERROR)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(GETGLOBALAA)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(ELSE IF)\b", const.KEYWORD, const.ELSE_IF, const.SPECIAL_INDENTATION],

    [r"(END IF)\b", const.KEYWORD, const.END_IF_TOKEN, const.DECREMENT_INDENTATION],
    [r"(EXIT FOR)\b", const.KEYWORD, const.EXIT, const.DECREMENT_INDENTATION],
    [r"(FOR EACH)\b", const.KEYWORD, const.FOR_EACH, const.INCREMENT_INDENTATION],
    [r"(END FOR)\b", const.KEYWORD, const.END_FOR_TOKEN, const.DECREMENT_INDENTATION],
    [r"(END WHILE)\b", const.KEYWORD, const.END_WHILE_TOKEN, const.DECREMENT_INDENTATION],
    [r"(EXIT WHILE)\b", const.KEYWORD, const.EXIT, const.DECREMENT_INDENTATION],
    [r"(END FUNCTION)\b", const.KEYWORD, const.END_FUNCTION_TOKEN, const.DECREMENT_INDENTATION],
    [r"(END SUB)\b", const.KEYWORD, const.END_SUB_TOKEN, const.DECREMENT_INDENTATION],
    [r"(LINE_NUM)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(GOTO)\b", const.KEYWORD, const.GOTO, const.NO_INDENTATION],
    [r"(CREATEOBJECT)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(ELSEIF)\b", const.KEYWORD, const.ELSE_IF, const.SPECIAL_INDENTATION],
    [r"(ENDIF)\b", const.KEYWORD, const.END_IF_TOKEN, const.DECREMENT_INDENTATION],
    [r"(ENDFOR)\b", const.KEYWORD, const.END_FOR_TOKEN, const.DECREMENT_INDENTATION],
    [r"(ENDWHILE)\b", const.KEYWORD, const.END_WHILE_TOKEN, const.DECREMENT_INDENTATION],
    [r"(EXITWHILE)\b", const.KEYWORD, const.EXIT, const.DECREMENT_INDENTATION],
    [r"(EXITFOR)\b", const.KEYWORD, const.EXIT, const.DECREMENT_INDENTATION],
    [r"(ENDFUNCTION)\b", const.KEYWORD, const.END_FUNCTION_TOKEN, const.DECREMENT_INDENTATION],
    [r"(ENDSUB)\b", const.KEYWORD, const.END_SUB_TOKEN, const.DECREMENT_INDENTATION],
    [r"(OBJFUN)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],

    [r"(Tan)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Sqr)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Sin)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Sgn)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Log)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Int)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Fix)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Exp)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Cint)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Cdbl)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Csng)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Cos)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Atn)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Abs)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Substitute)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(Tr)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(Val)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Stringi)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(Stri)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(Str)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(Right)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(Len)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Left)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(Instr)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(Chr)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Asc)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(LCase)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(Ucase)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(FormatJson)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(ParseJson)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(RunGarbageCollector)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(strtoi)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(FormatDrive)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(CreateDirectory)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(DeleteDirectory)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(DeleteFile)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(MatchFiles)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(MoveFile)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(CopyFile)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(WriteAsciiFile)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(ReadAsciiFile)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(RebootSystem)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(UpTime)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(GetInterface)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(Wait)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Sleep)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Eval)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(Run)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(Box)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(FindMemberFunction)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(Rnd)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(Type)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Void)\b", const.KEYWORD, const.TYPE, const.NO_INDENTATION],
    [r"(Dynamic)\b", const.KEYWORD, const.TYPE, const.NO_INDENTATION],
    [r"(Float)\b", const.KEYWORD, const.TYPE, const.NO_INDENTATION],
    [r"(Double)\b", const.KEYWORD, const.TYPE, const.NO_INDENTATION],
    [r"(Integer)\b", const.KEYWORD, const.TYPE, const.NO_INDENTATION],
    [r"(LongInteger)\b", const.KEYWORD, const.TYPE, const.NO_INDENTATION],
    [r"(String)\b", const.KEYWORD, const.TYPE, const.NO_INDENTATION],
    [r"(Boolean)\b", const.KEYWORD, const.TYPE, const.NO_INDENTATION],
    [r"(Object)\b", const.KEYWORD, const.TYPE, const.NO_INDENTATION],
    [r"(Until)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(On)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(Library)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(Generates)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(Implements)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(Event)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(Interface)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(Component)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],

    [r"(IF)\b", const.KEYWORD, const.IF, const.INCREMENT_INDENTATION],
    [r"(THEN)\b", const.KEYWORD, const.THEN, const.INCREMENT_INDENTATION],
    [r"(ELSE)\b", const.KEYWORD, const.ELSE, const.SPECIAL_INDENTATION],
    [r"(FOR)\b", const.KEYWORD, const.FOR, const.INCREMENT_INDENTATION],
    [r"(TO)\b", const.KEYWORD, const.TO, const.NO_INDENTATION],
    [r"(STEP)\b", const.KEYWORD, const.STEP, const.NO_INDENTATION],
    [r"(INVALID)\b", const.KEYWORD, const.VALUE, const.NO_INDENTATION],
    [r"(IN)\b", const.KEYWORD, const.IN, const.NO_INDENTATION],
    [r"(WHILE)\b", const.KEYWORD, const.WHILE, const.INCREMENT_INDENTATION],
    [r"(FUNCTION)\b", const.KEYWORD, const.FUNCTION, const.INCREMENT_INDENTATION],
    [r"(AS)\b", const.KEYWORD, const.AS, const.NO_INDENTATION],
    [r"(RETURN)\b", const.KEYWORD, const.RETURN, const.NO_INDENTATION],
    [r"(PRINT)\b", const.PRINT_KEYWORD, const.PRINT_KEYWORD, const.NO_INDENTATION],
    [r"(DIM)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(END)\b", const.KEYWORD, const.END_TOKEN, const.DECREMENT_INDENTATION],
    [r"(MOD)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(STOP)\b", const.KEYWORD, const.BUILT_IN_FUNCTION, const.NO_INDENTATION],
    [r"(AND)\b", const.KEYWORD, const.AND, const.NO_INDENTATION],
    [r"(BOX)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(EACH)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(EVAL)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(EXIT)\b", const.KEYWORD, const.EXIT, const.DECREMENT_INDENTATION],
    [r"(FALSE)\b", const.KEYWORD, const.VALUE, const.NO_INDENTATION],
    [r"(LET)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(NEXT)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(NOT)\b", const.KEYWORD, const.NOT, const.NO_INDENTATION],
    [r"(OR)\b", const.KEYWORD, const.OR, const.NO_INDENTATION],
    [r"(POS)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"('|rem) *(BSLINT)_(?P<command>[a-z_]+)(:(?P<param>[0-9]+))?.*", const.BSLINT_COMMAND,
     const.BSLINT_COMMAND, const.NO_INDENTATION],
    [r"('|REM\b) *(.*)", const.COMMENT, const.COMMENT, const.NO_INDENTATION],
    [r"(RUN)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(SUB)\b", const.KEYWORD, const.SUB, const.INCREMENT_INDENTATION],
    [r"(TAB)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],
    [r"(TYPE)\b", const.KEYWORD, const.KEYWORD, const.NO_INDENTATION],

    [r"\?", const.PRINT_KEYWORD, const.PRINT_KEYWORD, const.NO_INDENTATION],

    [r"(?P<value>^[a-z_][a-z0-9_]*)(?P<type>\$|%|!|#|&?)", const.ID, const.ID,
     const.NO_INDENTATION],
    ['\"(?P<value>.*?)\"', const.STRING, const.VALUE, const.NO_INDENTATION],
    [r"^\d*(\.?\d+){1}", const.NUMERIC, const.VALUE, const.NO_INDENTATION]
]

REGEXS = []
for regex in REGEX_LIST:
    REGEXS.append(Regex._make(regex))



# -*- coding: utf-8 -*-
"""
    BSLint
    ~~~~~
    A linter for the BrightScript language.
    :license: BSD, see LICENSE for more details.
"""

# pylint: disable=wildcard-import

from bslint.lexer.commands import *
from bslint.lexer.handlers.regex_handler import find_match
from .bslint import *
from .utilities.config_loader import *



"""bslint.bslint: provides entry point main()."""
import sys
import os
from bslint.interface_handler import InterfaceHandler as InterfaceHandler

__version__ = "0.6.1"


def main():
    try:
        interface_handler = InterfaceHandler()
        interface_handler.main()
    except KeyboardInterrupt:
        pass


def runner(to_lex=None):
    sys.argv = [sys.argv[0]]
    if to_lex is not None:
        sys.argv.append(os.path.abspath(to_lex))
    else:
        sys.argv.append(os.getcwd())
    interface_handler = InterfaceHandler()
    interface_handler.main()
    return interface_handler



import json
import sys
import bslint.constants as const


from filepaths import DEFAULT_CONFIG_FILE_PATH
CONFIG = ""


def load_config_file(user_filepath=None, default_filepath=DEFAULT_CONFIG_FILE_PATH, out=sys.stdout):
    default_json = get_default_config(default_filepath)
    try:
        user_json = get_user_config(user_filepath)
        default_json = overwrite_default_config(default_json, user_json)

    except FileNotFoundError:
        out.write(const.ERROR_COLOUR + "Cannot find bslintrc, using default config." + const.END_COLOUR)
    except ValueError:
        out.write(const.ERROR_COLOUR + "Unable to parse JSON in .bslintrc file" + const.END_COLOUR)
        default_json = None
    finally:
        global CONFIG
        CONFIG = default_json
        return default_json


def get_default_config(default_filepath):
    return read_json(default_filepath)


def get_user_config(user_filepath):
    user_json = ""
    if user_filepath is not None:
        user_json = read_json(user_filepath)
    return user_json


def overwrite_default_config(default_json, user_json):
    for setting in user_json:
        default_json[setting] = user_json[setting]
    return default_json


def read_json(filepath):
    with open(filepath, 'r') as file:
        return json.loads(file.read())



# pylint: disable=too-few-public-methods
class Token:
    # pylint: disable=too-many-arguments
    def __init__(self, group, lex_type, parse_type, line_number=None, token_type=None):
        self.group = group
        self.lexer_type = lex_type
        self.parser_type = parse_type
        self.token_type = token_type
        self.line_number = line_number

    def __eq__(self, other):
        return self.__dict__ == other.__dict__
